# -*- encoding: utf-8 -*-
# Cheqlist v0.3.0
# A simple Qt checklist.
# Copyright © 2015-2016, Chris Warrick.
# See /LICENSE for licensing information.

"""
The Cheqlist app.

:Copyright: © 2015-2016, Chris Warrick.
:License: BSD (see /LICENSE).
"""

from PyQt5 import QtWidgets

__all__ = ('UndoWidget',)


class UndoWidget(QtWidgets.QUndoView):
    """Undo view with checkbox setting."""

    _mw = None

    def showEvent(self, event):
        """Check the action associated with this window."""
        self._mw.actionShowUndoWindow.setChecked(True)

    def hideEvent(self, event):
        """Uncheck the action associated with this window."""
        self._mw.actionShowUndoWindow.setChecked(False)




# -*- encoding: utf-8 -*-
# Cheqlist v0.3.0
# A simple Qt checklist.
# Copyright © 2015-2016, Chris Warrick.
# See /LICENSE for licensing information.

"""
Utilties, mainly for parsing and file support.

:Copyright: © 2015-2016, Chris Warrick.
:License: BSD (see /LICENSE).
"""

import cheqlist
import re

SERIALIZE_FSTR = ' - [{x}] {markers}{text}{end_markers}\n'
zwnj = '\u200C'

RE_X = re.compile(r'\[([ *xX])\] ?(.*)')
RE_FMT = re.compile(r'^((?:\*|_|~~|<u>)*)(.*?)((?:\*|_|~~|</u>)*)$')
RE_BOLD = re.compile(r'(\*\*|__)(.*?)\1')
RE_ITALIC = re.compile(r'(\*|_)(.*?)\1')
RE_UNDERLINE = re.compile(r'(<u>)(.*?)(</u>)')
RE_STRIKEOUT = re.compile(r'(~~)(.*?)(~~)')


def _re_match(regex, line):
    """Match a regular expression and substitute it."""
    out = regex.sub(r'\2', line)
    return line != out, out


def parse_lines(lines):
    """Parse lines into entries."""
    for lineO in lines:
        line = lineO.strip()
        if line.startswith(('- ', '* ')):
            line = line[2:].strip()
        match = RE_X.match(line)
        if match:
            checked_text, line = match.groups()
            checked = checked_text != ' '
        else:
            checked = False

        bold = False
        italic = False
        underline = False
        strikeOut = False
        match = RE_FMT.match(line)
        if match:
            fmt_start, line, fmt_end = match.groups()
            fmt = fmt_start + 'f' + fmt_end
            bold, fmt = _re_match(RE_BOLD, fmt)
            italic, fmt = _re_match(RE_ITALIC, fmt)
            underline, fmt = _re_match(RE_UNDERLINE, fmt)
            strikeOut, fmt = _re_match(RE_STRIKEOUT, fmt)

            if fmt != 'f':
                cheqlist.log.warn(
                    "Failed to parse formats: %r left over from %r (%r; %d%d%d%d)",
                    fmt, lineO, line, bold, italic, underline, strikeOut)

            # Remove ZWNJ (see below)
            if line[0] == zwnj:
                line = line[1:-1]
            line = line.strip()

        yield (line, checked, bold, italic, underline, strikeOut)


def serialize_qt(items, log=True):
    """Serialize a Qt item list into GitHub Flavored Markdown."""
    done = 0
    for i in items:
        markers = ''
        x = 'x' if i.checkState() else ' '
        f = i.font()
        if f.bold():
            markers += '**'
        if f.italic():
            markers += '*'

        end_markers = markers

        if f.strikeOut():
            markers = '~~' + markers
            end_markers += '~~'
        if f.underline():
            markers = '<u>' + markers
            end_markers += '</u>'

        # Add ZWNJ to preserve non-markup text characters
        text = i.text()
        if text.startswith(('*', '~~', '<u>', '_')):
            text = zwnj + text + zwnj
        yield SERIALIZE_FSTR.format(x=x, markers=markers,
                                    end_markers=end_markers, text=text)
        done += 1
    if log:
        cheqlist.log.info("{0} tasks serialized".format(done))


def config_bool(value):
    """Represent a boolean in a way compatible with configparser."""
    return "true" if value else "false"



# -*- encoding: utf-8 -*-
# Cheqlist v0.3.0
# A simple Qt checklist.
# Copyright © 2015-2016, Chris Warrick.
# See /LICENSE for licensing information.

"""
The Cheqlist app.

:Copyright: © 2015-2016, Chris Warrick.
:License: BSD (see /LICENSE).
"""

from PyQt5 import QtCore, QtGui, QtWidgets

__all__ = ('PasteWindow',)


class PasteWindow(QtWidgets.QDialog):
    """A simple window to paste items."""

    _mw = None

    def __init__(self, main):
        """Create main window."""
        super(PasteWindow, self).__init__()
        self._mw = main

        self.gridLayout = QtWidgets.QGridLayout(self)
        self.pasteButton = QtWidgets.QPushButton(
            QtGui.QIcon.fromTheme("edit-paste"), "&Paste", self)
        self.clearButton = QtWidgets.QPushButton(
            QtGui.QIcon.fromTheme("edit-clear"), "&Clear", self)
        self.textBox = QtWidgets.QPlainTextEdit(self)
        self.label = QtWidgets.QLabel(
            "Every line below will be added to the list.\n"
            "GitHub Flavored Markdown is supported for basic markup "
            "(done status, <b>**bold**</b>, <i>*italic*</i>, "
            "<u>&lt;u&gt;underline&lt;/u&gt;</u>, <s>~~strikeout~~</s>).",
            self)
        self.label.setTextFormat(QtCore.Qt.RichText)
        self.label.setWordWrap(True)
        self.buttonBox = QtWidgets.QDialogButtonBox(self)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(
            QtWidgets.QDialogButtonBox.Ok |
            QtWidgets.QDialogButtonBox.Cancel)

        self.pasteButton.clicked.connect(self.paste)
        self.clearButton.clicked.connect(self.clear)

        self.gridLayout.addWidget(self.label, 0, 0, 2, 3)
        self.gridLayout.addWidget(self.pasteButton, 0, 3, 1, 1)
        self.gridLayout.addWidget(self.clearButton, 1, 3, 1, 1)
        self.gridLayout.addWidget(self.textBox, 2, 0, 1, 4)
        self.gridLayout.addWidget(self.buttonBox, 3, 0, 1, 4)

        self.setWindowTitle("Paste into Cheqlist")
        self.setWindowIcon(QtGui.QIcon.fromTheme("edit-paste"))
        self.setModal(True)
        self.resize(400, 400)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        QtCore.QMetaObject.connectSlotsByName(self)

        # Get clipboard contents
        self.paste()

    def paste(self, event=None):
        """Paste clipboard contents into the textbox."""
        self.textBox.clear()
        self.textBox.paste()

    def clear(self, event=None):
        """Clear the textbox."""
        self.textBox.clear()



# -*- encoding: utf-8 -*-
# Cheqlist v0.3.0
# A simple Qt checklist.
# Copyright © 2015-2016, Chris Warrick.
# See /LICENSE for licensing information.

"""
The Cheqlist app.

:Copyright: © 2015-2016, Chris Warrick.
:License: BSD (see /LICENSE).
"""

import os
import sys
import io
import time
import cheqlist
from cheqlist import utils, undocommands
from cheqlist.undowidget import UndoWidget
from cheqlist.pastewindow import PasteWindow
from PyQt5 import QtCore, QtGui, QtWidgets

__all__ = ('Main',)


class Main(QtWidgets.QMainWindow):
    """The main window of the app."""

    def __init__(self, app):
        """Create the GUI."""
        super(Main, self).__init__()
        self.app = app
        self.filename = None
        self.filebasename = None
        self.lastText = {}
        self.lastState = {}
        self.ignoreStruckOut = cheqlist.config.getboolean(
            'settings', 'ignore_struck_out')
        self.undoStack = QtWidgets.QUndoStack(self)

        self.centralwidget = QtWidgets.QWidget(self)
        self.verticalLayout = QtWidgets.QVBoxLayout(self.centralwidget)
        self.tasklist = QtWidgets.QListWidget(self.centralwidget)
        self.tasklist.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)
        self.tasklist.setDefaultDropAction(QtCore.Qt.MoveAction)

        self.verticalLayout.addWidget(self.tasklist)
        self.progressBar = QtWidgets.QProgressBar(self.centralwidget)
        self.progressBar.setTextVisible(True)
        self.progressBar.setFormat("%v/%m (%p%)")
        self.updateProgressBar()
        self.verticalLayout.addWidget(self.progressBar)
        self.setCentralWidget(self.centralwidget)
        self.toolBar = QtWidgets.QToolBar("Main", self)
        self.toolBar.setIconSize(QtCore.QSize(16, 16))
        self.toolBar.setFloatable(False)
        self.toolBar.setAllowedAreas(QtCore.Qt.TopToolBarArea)
        self.addToolBar(QtCore.Qt.TopToolBarArea, self.toolBar)
        self.editToolBar = QtWidgets.QToolBar("Edit", self)
        self.editToolBar.setIconSize(QtCore.QSize(16, 16))
        self.editToolBar.setFloatable(False)
        self.editToolBar.setAllowedAreas(QtCore.Qt.TopToolBarArea)
        self.addToolBar(QtCore.Qt.TopToolBarArea, self.editToolBar)
        self.insertToolBarBreak(self.editToolBar)
        menu = self.menuBar()
        self.fileMenu = menu.addMenu("&File")
        self.editMenu = menu.addMenu("&Edit")
        # self.helpMenu = menu.addMenu("&Help")

        self.actionAdd = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("list-add"), "&Add", self, shortcut='Ctrl+T',
            toolTip="Add", triggered=self.addItemHandler)

        self.actionEdit = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("edit-rename"), "&Edit", self,
            shortcut='Ctrl+E', toolTip="Edit", triggered=self.editItemHandler)

        self.actionDelete = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("list-remove"), "&Delete", self,
            shortcut='Delete', toolTip="Delete", triggered=self.delItemHandler)

        self.actionBold = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("format-text-bold"), "&Bold", self,
            shortcut='Ctrl+B', toolTip="Bold", checkable=True,
            triggered=self.boldItemHandler)

        self.actionItalic = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("format-text-italic"), "&Italic", self,
            shortcut='Ctrl+I', toolTip="Italic", checkable=True,
            triggered=self.italicItemHandler)

        self.actionUnderline = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("format-text-underline"), "&Underline", self,
            shortcut='Ctrl+U', toolTip="Underline", checkable=True,
            triggered=self.underlineItemHandler)

        self.actionStrikeOut = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("format-text-strikethrough"), "&Strike Out",
            self, shortcut='Ctrl+Alt+S', toolTip="Strike Out", checkable=True,
            triggered=self.strikeOutItemHandler)

        self.actionOpen = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("document-open"), "&Open…", self,
            shortcut='Ctrl+O', toolTip="Open", triggered=self.openHandler)

        self.actionSave = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("document-save"), "&Save…", self,
            shortcut='Ctrl+S', toolTip="Save", triggered=self.saveHandler)

        self.actionSaveAs = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("document-save-as"), "S&ave as…", self,
            shortcut='Ctrl+Shift+S', toolTip="Save as",
            triggered=self.saveAsHandler)

        self.actionClear = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("edit-clear-list"), "Clea&r", self,
            shortcut='Ctrl+R', toolTip="Clear", triggered=self.clear)

        self.actionPasteItems = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("edit-paste"), "&Paste items…", self,
            shortcut='Ctrl+V', toolTip="Paste items from clipboard",
            triggered=self.pasteItems)

        self.actionQuit = QtWidgets.QAction(
            QtGui.QIcon.fromTheme("application-exit"), "&Quit", self,
            shortcut='Ctrl+Q', toolTip="Quit", triggered=self.quit)

        self.actionCheckAll = QtWidgets.QAction(
            "Check &All", self, toolTip="Check all items in the list.",
            triggered=self.checkAll)

        self.actionCheckNone = QtWidgets.QAction(
            "&Uncheck All", self, toolTip="Uncheck all items in the list.",
            triggered=self.checkNone)

        self.actionCheckInvert = QtWidgets.QAction(
            "In&vert Selection", self, toolTip="Check all unchecked items "
            "and uncheck all checked items.", triggered=self.checkInvert)

        self.actionIgnoreStruckOut = QtWidgets.QAction(
            "I&gnore struck out tasks", self, toolTip="Ignore tasks that are "
            "struck out from counts", triggered=self.ignoreStruckOutHandler,
            checkable=True)

        self.actionUndo = self.undoStack.createUndoAction(self)
        self.actionUndo.setIcon(QtGui.QIcon.fromTheme("edit-undo"))
        self.actionUndo.setShortcut("Ctrl+Z")
        self.actionRedo = self.undoStack.createRedoAction(self)
        self.actionRedo.setIcon(QtGui.QIcon.fromTheme("edit-redo"))
        self.actionRedo.setShortcut("Ctrl+Y")

        self.actionShowUndoWindow = QtWidgets.QAction(
            "Show undo &window", self, toolTip="Show undo window with all "
            "operations performed", triggered=self.showUndoWindowHandler,
            checkable=True)

        self.toolBar.addAction(self.actionAdd)
        self.toolBar.addAction(self.actionDelete)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionOpen)
        self.toolBar.addAction(self.actionSave)
        self.toolBar.addAction(self.actionSaveAs)
        self.toolBar.addAction(self.actionClear)
        self.toolBar.addAction(self.actionQuit)

        self.editToolBar.addAction(self.actionBold)
        self.editToolBar.addAction(self.actionItalic)
        self.editToolBar.addAction(self.actionUnderline)
        self.editToolBar.addAction(self.actionStrikeOut)
        self.editToolBar.addSeparator()
        self.editToolBar.addAction(self.actionUndo)
        self.editToolBar.addAction(self.actionRedo)

        self.fileMenu.addAction(self.actionOpen)
        self.fileMenu.addAction(self.actionSave)
        self.fileMenu.addAction(self.actionSaveAs)
        self.fileMenu.addSeparator()
        self.fileMenu.addAction(self.actionQuit)

        self.editMenu.addAction(self.actionAdd)
        self.editMenu.addAction(self.actionEdit)
        self.editMenu.addAction(self.actionDelete)
        self.editMenu.addAction(self.actionClear)
        self.editMenu.addAction(self.actionPasteItems)
        self.editMenu.addSeparator()
        self.editMenu.addAction(self.actionBold)
        self.editMenu.addAction(self.actionItalic)
        self.editMenu.addAction(self.actionUnderline)
        self.editMenu.addAction(self.actionStrikeOut)
        self.editMenu.addSeparator()
        self.editMenu.addAction(self.actionUndo)
        self.editMenu.addAction(self.actionRedo)
        self.editMenu.addAction(self.actionShowUndoWindow)
        self.editMenu.addSeparator()
        self.editMenu.addAction(self.actionCheckAll)
        self.editMenu.addAction(self.actionCheckNone)
        self.editMenu.addAction(self.actionCheckInvert)
        self.editMenu.addSeparator()
        self.editMenu.addAction(self.actionIgnoreStruckOut)

        self.tasklist.itemChanged.connect(self.itemChangedHandler)
        self.tasklist.itemSelectionChanged.connect(self.selectionHandler)
        self.actionIgnoreStruckOut.setChecked(self.ignoreStruckOut)
        self.tasklist.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.tasklist.customContextMenuRequested.connect(
            self.tasklistMenuHandler)

        self.undoWidget = UndoWidget(self.undoStack)
        self.undoWidget.setWindowTitle("Operations — Cheqlist")
        self.undoWidget._mw = self

        self.setWindowIcon(QtGui.QIcon.fromTheme("checkbox"))
        self.updateWindowTitle()
        self.resize(250, 1000)
        self.updateUI()

        for a in sys.argv[1:]:
            self.readFile(a, clear=False)

        QtCore.QMetaObject.connectSlotsByName(self)
        self.show()

        cheqlist.log.info("Startup finished in {0:0.2f} s".format(
                          time.time() - cheqlist._starttime))

    # Item handling
    def items(self):
        """Yield items in the task list."""
        for n in range(0, self.tasklist.count()):
            yield self.tasklist.item(n)

    def clear(self, event=None, undoable=True):
        """Clear the task list."""
        if undoable:
            self.undoStack.push(undocommands.CommandClear(self))
        else:
            self.tasklist.clear()
            self.undoStack.clear()
            self.lastText = {}
            self.lastState = {}
        cheqlist.log.info("List cleared")
        self.updateUI()

    def addItem(self, text="New item", edit=False, checked=False, bold=False,
                italic=False, underline=False, strikeOut=False, undoable=True):
        """Add an item to the task list."""
        item = QtWidgets.QListWidgetItem(text, self.tasklist)
        if not edit:
            self.lastText[id(item)] = text
        item.setFlags(QtCore.Qt.ItemIsSelectable |
                      QtCore.Qt.ItemIsEditable |
                      QtCore.Qt.ItemIsDragEnabled |
                      QtCore.Qt.ItemIsDropEnabled |
                      QtCore.Qt.ItemIsUserCheckable |
                      QtCore.Qt.ItemIsEnabled)
        if checked:
            item.setCheckState(QtCore.Qt.Checked)
        else:
            item.setCheckState(QtCore.Qt.Unchecked)
        self.lastState[id(item)] = item.checkState()
        f = item.font()
        if bold:
            f.setBold(True)
        if italic:
            f.setItalic(True)
        if underline:
            f.setUnderline(True)
        if strikeOut:
            f.setStrikeOut(True)
        item.setFont(f)
        if edit:
            self.tasklist.addItem(item)
            item.setSelected(True)
            self.tasklist.editItem(item)
        elif undoable:
            self.undoStack.push(undocommands.CommandAdd(item, self))
        else:
            self.tasklist.addItem(item)
        self.updateUI()
        return self.tasklist.row(item)

    # File handling
    def loadFromText(self, items, undoable=False):
        """Load items from a text file."""
        self.updateUI_disable()
        for (item, checked, bold, italic, underline,
             strikeOut) in utils.parse_lines(items):
            self.addItem(item, False, checked, bold, italic, underline,
                         strikeOut, undoable)

        self.updateUI_enable()
        cheqlist.log.info("{0} tasks loaded".format(len(items)))

    # Action handling
    def addItemHandler(self, event):
        """Add an empty item."""
        self.addItem('', True)

    def editItemHandler(self, event):
        """Edit the currently selected item."""
        for i in self.tasklist.selectedItems():
            self.tasklist.editItem(i)

    def delItemHandler(self, event):
        """Delete the currently selected item."""
        for i in self.tasklist.selectedItems():
            self.undoStack.push(undocommands.CommandDelete(i, self))
        self.updateUI()

    def updateProgressBar(self):
        """Update progress bar."""
        done = 0
        count = self.tasklist.count()
        if self.ignoreStruckOut:
            for i in self.items():
                if i.font().strikeOut():
                    count -= 1
                elif i.checkState():
                    done += 1
        else:
            for i in self.items():
                if i.checkState():
                    done += 1
        self.progressBar.setMaximum(count)
        self.progressBar.setValue(done)

    def _formatItemHandler(self, attrName, commandName):
        for i in self.tasklist.selectedItems():
            if getattr(i.font(), attrName)():
                className = 'CommandUn' + commandName
            else:
                className = 'Command' + commandName
            self.undoStack.push(getattr(undocommands, className)(i, self))

    def boldItemHandler(self, event):
        """Toggle bold on an item."""
        self._formatItemHandler('bold', 'Bold')

    def italicItemHandler(self, event):
        """Toggle italic on an item."""
        self._formatItemHandler('italic', 'Italic')

    def underlineItemHandler(self, event):
        """Toggle underline on an item."""
        self._formatItemHandler('underline', 'Underline')

    def strikeOutItemHandler(self, event):
        """Toggle strike out on an item."""
        self._formatItemHandler('strikeOut', 'StrikeOut')

    def openHandler(self, event):
        """Open a file."""
        # Ask for unsaved changes first
        if not self.unsavedChanges():
            return

        openmode = cheqlist.config.get('directories', 'open_from')
        lastdir = cheqlist.config.get('directories', 'lastdir')
        path = os.path.expanduser(cheqlist.config.get('directories', openmode))
        fname, _ = QtWidgets.QFileDialog.getOpenFileName(
            self, "Open", path,
            "Markdown checklist files (*.cheqlist *.checklist *.md "
            "*.mdown *.markdown)")

        if not fname:
            return

        newpath = os.path.dirname(fname)
        if newpath != lastdir:
            cheqlist.config.set('directories', 'lastdir', newpath)
            cheqlist.config_write()

        self.readFile(fname)

    def readFile(self, fname, clear=True):
        """Read a file and load it."""
        cheqlist.log.info("Opening file " + fname)
        if clear:
            self.clear(undoable=False)
        with io.open(fname, 'r', encoding='utf-8') as fh:
            self.loadFromText(fh.readlines())
        self.filename = fname
        self.filebasename = os.path.basename(self.filename)
        self.updateWindowTitle()

    def saveHandler(self, event):
        """Save a file."""
        if self.filename:
            self.writeFile(self.filename)
        else:
            self.saveAsHandler(event)

    def saveAsHandler(self, event):
        """Save as a new file."""
        openmode = cheqlist.config.get('directories', 'open_from')
        lastdir = cheqlist.config.get('directories', 'lastdir')
        path = os.path.expanduser(cheqlist.config.get('directories', openmode))
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save", path,
            "Markdown checklist files (*.cheqlist *.checklist *.md "
            "*.mdown *.markdown)")

        if not fname:
            return

        newpath = os.path.dirname(fname)
        if newpath != lastdir:
            cheqlist.config.set('directories', 'lastdir', newpath)
            cheqlist.config_write()

        self.writeFile(fname)

    def writeFile(self, fname):
        """Serialize and write into a file."""
        cheqlist.log.info("Saving to file " + fname)
        with io.open(fname, 'w', encoding='utf-8') as fh:
            fh.writelines(utils.serialize_qt(self.items()))
        self.filename = fname
        self.filebasename = os.path.basename(fname)
        self.undoStack.clear()
        self.updateUI()

    def showUndoWindowHandler(self, event=None):
        """Show/hide the Undo window."""
        self.undoWidget.setVisible(not self.undoWidget.isVisible())

    def quit(self, event=None):
        """Display a message on quit via Ctrl+Q."""
        if self.unsavedChanges():
            cheqlist.log.info("*** Goodbye!")
            QtWidgets.qApp.quit()
        else:
            cheqlist.log.info("Aborted quit")

    def closeEvent(self, event=None):
        """Display a message on quit via the X button."""
        if self.unsavedChanges():
            cheqlist.log.info("*** Goodbye!")
            # super(Main, self).closeEvent(event)
            self.undoWidget.hide()
            event.accept()
            QtWidgets.qApp.quit()
        else:
            cheqlist.log.info("Aborted quit")
            event.ignore()

    def unsavedChanges(self):
        """If there are unsaved changes, ask the user if they want to save."""
        if not self.undoStack.isClean():
            if self.filename:
                msg = "List \"{0}\" has been modified.".format(
                    self.filebasename)
            else:
                msg = "The list has been modified."
            msgBox = QtWidgets.QMessageBox(self)
            msgBox.setWindowTitle("Cheqlist")
            msgBox.setText(msg)
            msgBox.setInformativeText("Do you want to save your changes?")
            msgBox.setIcon(QtWidgets.QMessageBox.Question)
            msgBox.setStandardButtons(QtWidgets.QMessageBox.Save |
                                      QtWidgets.QMessageBox.Discard |
                                      QtWidgets.QMessageBox.Cancel)
            msgBox.setDefaultButton(QtWidgets.QMessageBox.Save)
            ret = msgBox.exec()
            if ret == QtWidgets.QMessageBox.Save:
                self.saveHandler(None)
                return True
            elif ret == QtWidgets.QMessageBox.Discard:
                return True
            elif ret == QtWidgets.QMessageBox.Cancel:
                return False
        else:
            return True

    def checkAll(self, event=None):
        """Check all items (complete the list)."""
        self.undoStack.push(undocommands.CommandCheckAll(self))

    def checkNone(self, event=None):
        """Uncheck all items (reset the list)."""
        self.undoStack.push(undocommands.CommandCheckNone(self))

    def checkInvert(self, event=None):
        """Check all unchecked items, uncheck all checked items."""
        self.undoStack.push(undocommands.CommandCheckInvert(self))

    def ignoreStruckOutHandler(self, event=None):
        """Toggle the setting that ignores struck out events."""
        self.ignoreStruckOut = self.actionIgnoreStruckOut.isChecked()
        cheqlist.config.set("settings", "ignore_struck_out",
                            utils.config_bool(self.ignoreStruckOut))
        cheqlist.config_write()
        self.updateProgressBar()

    def pasteItems(self, event=None):
        """Open the paste items dialog box."""
        pw = PasteWindow(self)
        if pw.exec():
            self.loadFromText(pw.textBox.toPlainText().splitlines())

    # UI functions and helpers
    def tasklistMenuHandler(self, point):
        """Show the right-click menu of a task list."""
        pos = self.tasklist.mapToGlobal(point)
        m = QtWidgets.QMenu()
        if self.tasklist.count() == 0:
            m.addAction(self.actionAdd)
        else:
            m.addAction(self.actionEdit)
            m.addAction(self.actionDelete)
            m.addAction(self.actionBold)
            m.addAction(self.actionItalic)
            m.addAction(self.actionUnderline)
            m.addAction(self.actionStrikeOut)
        m.exec(pos)

    def cleanChanged(self, clean):
        """Update the window title to reflect that the window is dirty."""
        self.updateWindowTitle()

    def updateWindowTitle(self):
        """Set the appropriate window title (file name, dirty status)."""
        if self.filename and self.undoStack.isClean():
            self.setWindowTitle("{0} — Cheqlist".format(self.filebasename))
        elif self.filename:
            self.setWindowTitle("{0} * — Cheqlist".format(self.filebasename))
        else:
            self.setWindowTitle("Cheqlist")

    def itemChangedHandler(self, item):
        """Handle item changes."""
        try:
            lt = self.lastText[id(item)]
            it = item.text()
            if lt is None and it:
                # Text added
                self.undoStack.push(undocommands.CommandAdd(item, self))
            elif lt is not None and it != lt:
                # Item text changed, let’s make this an edit action
                self.undoStack.push(undocommands.CommandEdit(
                    item, self, lt, it))
        except KeyError:
            self.lastText[id(item)] = None

        cs = item.checkState()
        try:
            ls = self.lastState[id(item)]
            if ls != cs and cs == QtCore.Qt.Checked:
                self.undoStack.push(undocommands.CommandCheck(item, self))
            elif ls != cs and cs == QtCore.Qt.Unchecked:
                self.undoStack.push(undocommands.CommandUnCheck(item, self))
        except KeyError:
            # lastState is set only after checkState is set
            pass
        self.updateUI()

    def updateBoldAction(self):
        """Set the bold action check status."""
        s = False
        for i in self.tasklist.selectedItems():
            s = i.font().bold()

        self.actionBold.setChecked(s)

    def updateItalicAction(self):
        """Set the italic action check status."""
        s = False
        for i in self.tasklist.selectedItems():
            s = i.font().italic()

        self.actionItalic.setChecked(s)

    def updateUnderlineAction(self):
        """Set the underline action check status."""
        s = False
        for i in self.tasklist.selectedItems():
            s = i.font().underline()

        self.actionUnderline.setChecked(s)

    def updateStrikeOutAction(self):
        """Set the strike out action check status."""
        s = False
        for i in self.tasklist.selectedItems():
            s = i.font().strikeOut()

        self.actionStrikeOut.setChecked(s)

    def updateDisabledButtons(self):
        """Disable buttons if the list is empty."""
        if self.tasklist.count() == 0:
            self.actionDelete.setEnabled(False)
            self.actionBold.setEnabled(False)
            self.actionBold.setChecked(False)
            self.actionItalic.setEnabled(False)
            self.actionItalic.setChecked(False)
            self.actionUnderline.setEnabled(False)
            self.actionUnderline.setChecked(False)
            self.actionStrikeOut.setEnabled(False)
            self.actionStrikeOut.setChecked(False)
            self.progressBar.setEnabled(False)
        else:
            self.actionDelete.setEnabled(True)
            self.actionBold.setEnabled(True)
            self.actionItalic.setEnabled(True)
            self.actionUnderline.setEnabled(True)
            self.actionStrikeOut.setEnabled(True)
            self.progressBar.setEnabled(True)

    def updateUI(self):
        """Update all UI."""
        self.updateProgressBar()
        self.selectionHandler()
        self.updateDisabledButtons()
        self.updateWindowTitle()

    def _updateUI_disabled(self):
        """Do nothing."""
        pass

    _updateUI_enabled = updateUI

    def updateUI_disable(self):
        """Disable the updateUI function for long running operations."""
        self.progressBar.setMaximum(0)
        self.progressBar.setValue(0)
        self.updateUI = self._updateUI_disabled

    def updateUI_enable(self):
        """Enable the updateUI function after long running operations."""
        self.updateUI = self._updateUI_enabled
        self.updateUI()

    def selectionHandler(self):
        """Update actions when the selection changes."""
        self.updateBoldAction()
        self.updateItalicAction()
        self.updateUnderlineAction()
        self.updateStrikeOutAction()



# -*- encoding: utf-8 -*-
# Cheqlist v0.3.0
# A simple Qt checklist.
# Copyright © 2015-2016, Chris Warrick.
# See /LICENSE for licensing information.

"""
Undo commands for Cheqlist.

:Copyright: © 2015-2016, Chris Warrick.
:License: BSD (see /LICENSE).
"""

from PyQt5.QtWidgets import QUndoCommand
from PyQt5 import QtCore

__all__ = ('CommandAdd', 'CommandDelete', 'CommandClear', 'CommandCheck',
           'CommandUnCheck', 'CommandBold', 'CommandUnBold', 'CommandItalic',
           'CommandUnItalic', 'CommandUnderline', 'CommandUnUnderline',
           'CommandStrikeOut', 'CommandUnStrikeOut')


class CheqlistUndoCommand(QUndoCommand):
    """Generic Undo command."""

    prefix = ""

    def __init__(self, item, window):
        """Create the action."""
        self.item = item
        self.window = window
        self.desc = ' '.join((self.prefix, item.text()))
        super(CheqlistUndoCommand, self).__init__(self.desc)


class CommandAdd(CheqlistUndoCommand):
    """Add a task."""

    prefix = "Add"

    def undo(self):
        """Undo the command."""
        self.window.tasklist.takeItem(self.window.tasklist.row(self.item))
        del self.window.lastText[id(self.item)]

    def redo(self):
        """Redo the command."""
        self.window.tasklist.addItem(self.item)
        self.window.lastText[id(self.item)] = self.item.text()


class CommandDelete(CheqlistUndoCommand):
    """Delete a task."""

    prefix = "Delete"

    def undo(self):
        """Undo the command."""
        self.window.tasklist.addItem(self.item)
        self.window.lastText[id(self.item)] = self.item.text()

    def redo(self):
        """Redo the command."""
        self.window.tasklist.takeItem(self.window.tasklist.row(self.item))
        del self.window.lastText[id(self.item)]


class CommandClear(QUndoCommand):
    """Clear the list."""

    desc = "Clear"

    def __init__(self, window):
        """Create the action."""
        self.window = window
        self.items = list(self.window.items())
        super(CommandClear, self).__init__(self.desc)

    def undo(self):
        """Undo the command."""
        for i in self.items:
            self.window.tasklist.addItem(i)

    def redo(self):
        """Redo the command."""
        for i in self.items:
            self.window.tasklist.takeItem(self.window.tasklist.row(i))


class CommandEdit(QUndoCommand):
    """Edit a task."""

    prefix = "Edit"

    def __init__(self, item, window, oldText, newText):
        """Create the action."""
        self.item = item
        self.window = window
        self.desc = ' '.join((self.prefix, newText))
        self.oldText = oldText
        self.newText = newText
        super(QUndoCommand, self).__init__(self.desc)

    def undo(self):
        """Undo the command."""
        self.window.lastText[id(self.item)] = self.oldText
        self.item.setText(self.oldText)

    def redo(self):
        """Redo the command."""
        self.window.lastText[id(self.item)] = self.newText
        self.item.setText(self.newText)


class CommandCheck(CheqlistUndoCommand):
    """Mark a task done."""

    prefix = "Check"
    undoValue = QtCore.Qt.Unchecked
    redoValue = QtCore.Qt.Checked

    def undo(self):
        """Undo the command."""
        self.window.lastState[id(self.item)] = self.undoValue
        self.item.setCheckState(self.undoValue)

    def redo(self):
        """Redo the command."""
        self.window.lastState[id(self.item)] = self.redoValue
        self.item.setCheckState(self.redoValue)


class CommandUnCheck(CommandCheck):
    """Mark a task undone."""

    prefix = "Uncheck"
    undoValue = QtCore.Qt.Checked
    redoValue = QtCore.Qt.Unchecked


class CommandFormat(CheqlistUndoCommand):
    """Generic formatting command (base class)."""

    undoValue = False
    functionName = ""

    def _changeFont(self, value):
        """Change the font to the desired formatting."""
        f = self.item.font()
        getattr(f, self.functionName)(value)
        self.item.setFont(f)

    def undo(self):
        """Undo the command."""
        self._changeFont(self.undoValue)

    def redo(self):
        """Redo the command."""
        self._changeFont(not self.undoValue)


class CommandBold(CommandFormat):
    """Make something bold."""

    prefix = "Bold"
    functionName = "setBold"


class CommandUnBold(CommandFormat):
    """Make something not bold."""

    prefix = "Remove Bold"
    functionName = "setBold"
    undoValue = True


class CommandItalic(CommandFormat):
    """Make something italic."""

    prefix = "Italicize"
    functionName = "setItalic"


class CommandUnItalic(CommandFormat):
    """Make something not italic."""

    prefix = "Deitalicize"
    functionName = "setItalic"
    undoValue = True


class CommandUnderline(CommandFormat):
    """Make something underlined."""

    prefix = "Underline"
    functionName = "setUnderline"


class CommandUnUnderline(CommandFormat):
    """Make something not underlined."""

    prefix = "Remove Underline"
    functionName = "setUnderline"
    undoValue = True


class CommandStrikeOut(CommandFormat):
    """Make something struck out."""

    prefix = "Strike Out"
    functionName = "setStrikeOut"

    def undo(self):
        """Undo the command."""
        self._changeFont(self.undoValue)
        if self.window.ignoreStruckOut:
            self.window.updateProgressBar()

    def redo(self):
        """Redo the command."""
        self._changeFont(not self.undoValue)
        if self.window.ignoreStruckOut:
            self.window.updateProgressBar()


class CommandUnStrikeOut(CommandStrikeOut):
    """Make something not struck out."""

    prefix = "Remove Strike Out"
    undoValue = True


class CommandPatternCheck(QUndoCommand):
    """Generic Check Pattern command."""

    desc = ""
    pattern = None

    def __init__(self, window):
        """Create the action."""
        self.window = window
        self.items = {}
        for i in self.window.items():
            self.items[id(i)] = i.checkState()
        super(CommandPatternCheck, self).__init__(self.desc)

    def undo(self):
        """Undo the command."""
        self.window.updateUI_disable()
        for i in self.window.items():
            self.window.lastState[id(i)] = self.items[id(i)]
            i.setCheckState(self.items[id(i)])
        self.window.updateUI_enable()

    def redo(self):
        """Redo the command."""
        self.window.updateUI_disable()
        for i in self.window.items():
            self.window.lastState[id(i)] = self.pattern
            i.setCheckState(self.pattern)
        self.window.updateUI_enable()


class CommandCheckAll(CommandPatternCheck):
    """Check all items."""

    pattern = QtCore.Qt.Checked
    desc = "Check All"


class CommandCheckNone(CommandPatternCheck):
    """Uncheck all items."""

    pattern = QtCore.Qt.Unchecked
    desc = "Uncheck All"


class CommandCheckInvert(CommandPatternCheck):
    """Check all items."""

    desc = "Invert Selection"

    def redo(self):
        """Redo the command."""
        self.window.updateUI_disable()
        for i in self.window.items():
            if self.items[id(i)] == QtCore.Qt.Checked:
                ns = QtCore.Qt.Unchecked
            else:
                ns = QtCore.Qt.Checked
            self.window.lastState[id(i)] = ns
            i.setCheckState(ns)
        self.window.updateUI_enable()



# -*- encoding: utf-8 -*-
# Cheqlist v0.3.0
# A simple Qt checklist.
# Copyright © 2015-2016, Chris Warrick.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the author of this software nor the names of
#    contributors to this software may be used to endorse or promote
#    products derived from this software without specific prior written
#    consent.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
A simple Qt checklist.

:Copyright: © 2015-2016, Chris Warrick.
:License: BSD (see /LICENSE).
"""

import os
import pkg_resources
import io
import logging
import time
try:
    import configparser
except ImportError:
    import ConfigParser as configparser

__title__ = 'Cheqlist'
__version__ = '0.3.0'
__author__ = 'Chris Warrick'
__license__ = '3-clause BSD'
__docformat__ = 'restructuredtext en'

__all__ = ('main',)

# Time
_starttime = time.time()

# import gettext
# G = gettext.translation('Cheqlist', '/usr/share/locale', fallback='C')
# _ = G.gettext

# Config directory setup
confhome = os.getenv('XDG_CONFIG_HOME')
if confhome is None:
    confhome = os.path.expanduser('~/.config/')

kwdir = os.path.join(confhome, 'kwpolska')
confdir = os.path.join(kwdir, 'cheqlist')
confpath = os.path.join(confdir, 'cheqlist.ini')
logpath = os.path.join(confdir, 'cheqlist.log')

if not os.path.exists(confhome):
    os.mkdir(confhome)

if not os.path.exists(kwdir):
    os.mkdir(kwdir)

if not os.path.exists(confdir):
    os.mkdir(confdir)

# Logging configuration
logging.basicConfig(format='%(asctime)-15s [%(levelname)-7s] '
                    ':%(name)-10s: %(message)s',
                    filename=logpath,
                    level=logging.DEBUG)
log = logging.getLogger('cheqlist')
console = logging.StreamHandler()
console.setLevel(logging.INFO)
console.setFormatter(logging.Formatter('[%(levelname)-7s] '
                     ':%(name)-10s: %(message)s'))
logging.getLogger('').addHandler(console)
log.info('*** Cheqlist v' + __version__)


if not os.path.exists(confpath):
    log.warn("cheqlist.ini does not exist, creating")
    with open(confpath, 'wb') as fh:
        fh.write(pkg_resources.resource_string(
            'cheqlist', 'data/cheqlist.ini.skel'))

# Configuration file support
config = configparser.ConfigParser()
config.read_string(pkg_resources.resource_string(
            'cheqlist', 'data/cheqlist.ini.skel').decode('utf-8'))
config.read([confpath], encoding='utf-8')
log.info("Read config from cheqlist.ini")


def config_write():
    with io.open(confpath, 'w', encoding='utf-8') as fh:
        config.write(fh)
    log.info("Config written to cheqlist.ini")



#!/usr/bin/env python3
import os
import pathlib
import random
import sys
MARKERS = [('*', '*'), ('**', '**'), ('_', '_'), ('__', '__'), ('~~', '~~'), ('<u>', '</u>')]
CHECKMARKS = ['', '- ', '-', '*', '* ', '- [', '- [ ]', '- [ ] ', '- [x]', '* [X]', '* [*]', '- [-]', '[ ]', '[X] ', '[x]', '[X]', '[*]', '     [x]', '    [ ]  ', '           ']
base = pathlib.Path("/usr/share/dict")

if len(sys.argv) != 3:
    print("usage: randomDemo.py entries outfile.cheqlist")
    exit(2)

entries = int(sys.argv[1])
outfile = sys.argv[2]

dictionaries = random.sample(os.listdir(str(base)), 3)
WORDS = ['', ' ', '    ']
for dictionary in dictionaries:
    with open(str(base.joinpath(dictionary)), 'r', encoding='utf-8') as fh:
        WORDS += fh.read().split()

WORDS = list(set(WORDS))

print("initialized with {0} words, {1} markers, {2} checkmarks".format(
    len(WORDS), len(MARKERS), len(CHECKMARKS)))
print("dictionaries:", ", ".join(dictionaries))

with open(outfile, 'w', encoding='utf-8') as fh:
    for _e in range(entries):
        # Text: 1-10 random words
        text = ' '.join(random.sample(WORDS, random.randint(1, 10)))
        # Checkmark: 1 random
        checkmark = random.choice(CHECKMARKS)
        # 25% chance of doubling checkmark
        # 15% chance of adding one more checkmark
        if random.random() >= 0.75:
            checkmark += ' ' + checkmark
        elif random.random() >= 0.85:
            checkmark += ' ' + random.choice(CHECKMARKS)
        markers_count = random.randint(0, 6)
        start_markers = end_markers = ''
        for s, e in random.sample(MARKERS, markers_count):
            start_markers = s + start_markers
            end_markers += e

        fh.write(checkmark + start_markers + text + end_markers + '\n')

print("written {0} entries".format(entries))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import io
from setuptools import setup

setup(name='cheqlist',
      version='0.3.0',
      description='A simple Qt checklist.',
      keywords='cheqlist',
      author='Chris Warrick',
      author_email='chris@chriswarrick.com',
      url='https://github.com/Kwpolska/cheqlist',
      license='3-clause BSD',
      long_description=io.open('./docs/README.rst', 'r', encoding='utf-8').read(),
      platforms='any',
      zip_safe=False,
      include_package_data=True,
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=['Development Status :: 3 - Alpha',
                   'Environment :: X11 Applications :: Qt',
                   'License :: OSI Approved :: BSD License',
                   'Topic :: Office/Business :: Scheduling',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.3',
                   'Programming Language :: Python :: 3.4',
                   'Programming Language :: Python :: 3.5'],
      packages=['cheqlist'],
      data_files=[('share/applications', ['freedesktop/cheqlist.desktop']),
                  ('share/mime/packages', ['freedesktop/cheqlist.xml']),],
      entry_points={
          'gui_scripts': [
              'cheqlist = cheqlist.__main__:main',
          ]
      },
      )



# -*- encoding: utf-8 -*-
# Cheqlist v0.3.0
# A simple Qt checklist.
# Copyright © 2015-2016, Chris Warrick.
# See /LICENSE for licensing information.

"""
Run the app.

:Copyright: © 2015-2016, Chris Warrick.
:License: BSD (see /LICENSE).
"""

from cheqlist import log, logpath
from cheqlist.app import QtWidgets, Main
import sys
import traceback

__all__ = ('main',)


def main():
    """The main routine for the UI."""
    try:
        app = QtWidgets.QApplication(sys.argv)
        main = Main(app)  # NOQA
        sys.exit(app.exec_())
    except Exception as e:
        log.exception(e)
        # tb ends with newline
        tb = ''.join(traceback.format_exception_only(type(e), e))
        QtWidgets.QMessageBox.critical(
            None, "Cheqlist", "An exception occurred:\n\n{0}\n"
            "Please report a bug in Cheqlist and attach recent log entries.\n"
            "The log is stored in {1}".format(tb, logpath))
        sys.exit(1)

if __name__ == '__main__':
    main()
